/* 
** DAVfsd protocol process 
** Copyright (C) 2001 SungHun Kim <hunkim@kebi.com>
**
** This library and program is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**   
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
** MA 02111-1307, USA
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/timeb.h>

#include <config.h>

#include <libdav.h>
#include <davfsd.h>
#include <xml_result.h>
#include <xmlparse.h>

extern char *cache_dir;

char *new_path(char *path, char *add_path) {
    char buf[BUF_SIZE];
    sprintf(buf, "%s%s", path, add_path);
    
    safe_free(path);
    return reduce_path(buf);
}

int do_xml(dav_config *dc, dav_rec *rec) {
    int done=0, to_read;
    int i, n, tn = 0;    
    char buf[BUF_SIZE]="";

    /* if xml parser is available */
    if(!rec->xml_p) {
	if(rec->hh.content_length && dc->connected) /* just read it and skip */
	    for(i=0;tn < rec->hh.content_length || !dc->connected ;i++, tn+=n) {
		to_read = min(rec->hh.content_length-tn, BUF_SIZE-1);
	    
		if((n=sock_read(dc->ns, buf, to_read)) <=0 ) 
		    break;
	    }
	/* connection closed and no content-length */
    } else if(!dc->connected && !rec->hh.has_content_length && !rec->hh.chunked) {
	while((n=read(dc->sock, buf, sizeof(buf)-1))>0) {
#ifdef XML_DEBUG
	    puts(buf);
#endif
	    if (! XML_Parse(rec->xml_p, buf, n, done)) {
		fprintf(stderr, "Parse error at line %d:\n%s\n",
			XML_GetCurrentLineNumber(rec->xml_p),
			XML_ErrorString(XML_GetErrorCode(rec->xml_p)));
	    }
	}
    } else {
   	for(i=0;tn < rec->hh.content_length || !dc->connected ;i++, tn+=n) { 
	    to_read = min(rec->hh.content_length-tn, BUF_SIZE-1);
	    
	    if((n=sock_read(dc->ns, buf, to_read)) <=0 ) 
		break;
#ifdef XML_DEBUG
	    puts(buf);
#endif
	    
	    if (! XML_Parse(rec->xml_p, buf, n, done)) {
		fprintf(stderr, "Parse error at line %d:\n%s\n",
			XML_GetCurrentLineNumber(rec->xml_p),
			XML_ErrorString(XML_GetErrorCode(rec->xml_p)));
	    }
	}
    }
    
    return tn;
}


int end_xml(dav_config *dc, dav_rec *rec) {
    int done=1;

    if(!rec->xml_p) 
	return -1;
    
    if(!rec->hh.content_length)
	return -1;

    if (! XML_Parse(rec->xml_p, NULL, 0, done)) {
	fprintf(stderr, "Parse2 error at line %d:\n%s\n",
		XML_GetCurrentLineNumber(rec->xml_p),
		XML_ErrorString(XML_GetErrorCode(rec->xml_p)));
	exit(-1);
    }
    
    return 0;
}

int dav_fill_finfo(xml_result *res, dav_config *dc, char *new_path, struct dav_finfo *finfo) {
    int i;
    dav_rec rec;
    char *last=NULL, *ptr;
   
    /* fill content size */
    if(!finfo->f_isdir)  {
	/* get content size */
	ptr = xr_search_tag(res, "getcontentlength");
	if(ptr) {
	    finfo->f_size = strtol(ptr, NULL, 10);
	} else { /* use HEAD method */
	    dav_rec_clear(&rec);
	    safe_free(dc->path);
	    dc->path = safe_strdup(new_path);
	    if(dav_head(dc, &rec)==200) 
		finfo->f_size = rec.hh.content_length;
	    dav_rec_free(&rec);
	}
    }

    /* get last one, it shoud be file name */
    for(i=0;;i++) {
	if(!i) 
	    ptr = strtok(new_path, "/\n\r ");
	else 
	    ptr = strtok(NULL, "/\n\r ");
	
	if(!ptr) break;
	last = ptr;
    }
    
    /* itself or invalid name */
    if(!last) last = "/";
    
    /* fill file name and length */
    strncpy(finfo->f_name, last, DAV_MAXNAMELEN-1);
    unescape_uri(finfo->f_name);
    finfo->f_name_len = strlen(finfo->f_name);
  
    /* fill time structure */
    finfo->f_ctime = dav_parse_date(xr_search_tag(res, "creationdate"));
    
    /* do it have mtime? */
    if((ptr=xr_search_tag(res, "davfsmtime"))) 
	finfo->f_mtime = strtol(ptr, NULL, 10);
    else 
        finfo->f_mtime = dav_parse_date(xr_search_tag(res, "getlastmodified"));
    
    finfo->f_atime = finfo->f_mtime;
    
    return i;
}
 
int dav_mk_res(char *buf, int ret, int count) {
	sprintf(buf, "%03d %014d", ret, count);
	return 0;
}

int client_unlock(dav_config *dc, char *lock_token) {
    int ret;
    dav_rec rec;
       
    if(!lock_token)
	return 0;

    dav_rec_clear(&rec);
    rec.lock_token = safe_strdup(lock_token);

    ret = dav_unlock(dc, &rec);

    dav_rec_free(&rec);
    return ret;
}   


/*
** lock 
** return lock_toekn and make lock_token
**
*/
int client_lock(dav_config *dc, char *lock_token) {
    int ret;
    xml_ctx c;    
    dav_rec rec;
    char *data;
    
    dav_rec_clear(&rec);

    rec.xml_p = XML_ParserCreate(NULL);
    if (! rec.xml_p) {
	fprintf(stderr, "Couldn't allocate memory for parser\n");
	exit(-1);
    }
    
    /* initialize xml_ctx */
    xr_init(&c);
    
    XML_SetUserData(rec.xml_p, &c);
    XML_SetElementHandler(rec.xml_p, xr_start, xr_end);
    XML_SetCharacterDataHandler(rec.xml_p, xr_data_handle);
    
    /* Assign xml parse function */
    rec.do_func  = do_xml;
    rec.end_func = end_xml;
    
    /* Make XML  for lock */
    rec.xml = safe_strdup("\
<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n\
<lockinfo xmlns=\"DAV:\">\n\
<locktype>\n\
<write/>\n\
</locktype>\n\
<lockscope>\n\
<exclusive/>\n\
</lockscope>\n\
<owner>DAVfs</owner>\n\
</lockinfo>\n\
");
    ret = dav_lock(dc, &rec);

    if(rec.xml_p)
        XML_ParserFree(rec.xml_p);
    
    /* read lock token from header */
    if(rec.hh.lock_token) {
	strcpy(lock_token, rec.hh.lock_token);
    } else {  /* If lock token is not available in head */ 
	data = xr_find_data(&c, 
			    "/prop/lockdiscovery/activelock/locktoken/href");
	if(data) {
	    if(strtok(data, "\n\r"))
		sprintf(lock_token, "<%s>", data);
	}else 
	    ret = 0;
    }
    
    xr_free(&c);
    dav_rec_free(&rec);
    
    return ret;
}

int client_attr_cache(dav_config *dc, struct dav_finfo *finfo) {
    int ret;
    char buf[BUF_SIZE];
    struct descriptor_data *d;

    /* get descritor data */
    d = (struct descriptor_data *)dc->userdata;
    if(!d) 
	return -1;
    
    dav_mk_res(buf, 200, sizeof(struct dav_finfo));
    sock_sendline(d->ns, buf);
	    
    /* send it to socket */
    if(sock_fullwrite(d->ns, (void*)finfo, sizeof(struct dav_finfo))<0) {
	perror("fatter write");
    }
    
    return ret;
}

int client_ls_attr(dav_config *dc, char *arg1, char *arg2, int depth) {
    int ret, count=0;
    xml_ctx c;    
    dav_rec rec;
    char *data;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    struct dav_finfo finfo;
    int i=0, path_len;
    char *oldpath=NULL;
    xml_result *one_res;

    dav_rec_clear(&rec);

    /* get descritor data */
    d = (struct descriptor_data *)dc->userdata;
    if(!d) 
	goto out;
	
    rec.xml_p = XML_ParserCreate(NULL);
    if (! rec.xml_p) {
	fprintf(stderr, "Couldn't allocate memory for parser\n");
	exit(-1);
    }
    
    /* initialize xml_ctx */
    xr_init(&c);
    
    XML_SetUserData(rec.xml_p, &c);
    XML_SetElementHandler(rec.xml_p, xr_start, xr_end);
    XML_SetCharacterDataHandler(rec.xml_p, xr_data_handle);
    
    /* depth should be 1 */
    rec.o.depth  = depth;
    rec.do_func  = do_xml;
    rec.end_func = end_xml;
    
    if(arg1) {
 	safe_free(dc->path);
	dc->path = reduce_path(arg1);
    }
    
    /* 
    ** We need only 
    **   - status
    **   - href
    **   - collection
    **   - getcontentlength
    **   - creationdate
    **   - davfsmtime
    **   - getlastmodified
    **  properties
    */
    rec.xml = safe_strdup("\
<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n\
<D:propfind xmlns:D=\"DAV:\">\n\
  <D:allprop/>
</D:propfind>\
");
/*
rec.xml = safe_strdup("\
<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n\
<D:propfind xmlns:D=\"DAV:\">\n\
<D:prop>\n\
<D:status/>\n\
<D:href/>\n\
<D:getcontentlength/>\n\
<D:davfsmtime/>\n\
<D:getlastmodified/>\n\
<D:creationdate/>\n\
<D:collection/>\n\
</D:prop>\n\
</D:propfind>\
");
*/

    ret = dav_propfind(dc, &rec);

    if(rec.xml_p)
        XML_ParserFree(rec.xml_p);

    /* depth is 2 because of /D:multistatus/D:response */
    count = xr_count(&c, 2);
 
    /* attr */	
    if(depth==0) {
        one_res = xr_one_result(&c, SEP_XML);
	if(one_res) {
	    data = xr_search_tag(one_res, "status");
	    if(data) {
		strcpy(buf, data);
		strtok(buf," \t\n\r");
		data = strtok(NULL," \t\n\r");
		if(data) 
		    ret = atoi(data);
	    }
        }
    }	

    dav_mk_res(buf, ret, count*sizeof(finfo));
    sock_sendline(d->ns, buf);

    /* keep old path to remove itself directory */
    oldpath = safe_strdup(dc->path);

    for(i=0;i<count;i++) {
	
	/* clear finfo memory */
	memset(&finfo, 0x00, sizeof(finfo));

	/* depth is 3 because of /D:multistatus/D:response */
	if(!i) {
	    one_res = xr_one_result(&c, SEP_XML);
	} else {
	    one_res = xr_one_result(NULL, SEP_XML);
	}

	if(one_res) {
	    data = xr_search_tag(one_res, "href");
#ifdef XML_DEBUG
	    xr_print_friend(one_res);
#endif
	    /* if full url */
	    if(!strncasecmp("http://", data, 7)) 
		data = strchr(data+7, '/');
	    if(xr_search_tag(one_res, "collection")) 
	        finfo.f_isdir = 1;
	}
	else 
	    data = NULL;

	/* we should send data */
	if(!data) strcpy(buf, "");
	else strcpy(buf, data);

	/* delete white spaces*/
        strtok(buf, "\r\n\t ");

	/* if dir and same as itself, just eleminate*/
	path_len = strlen(buf)-1;

	//finfo.f_isdir= (*(buf+path_len)=='/');

	if(finfo.f_isdir && !strncmp(oldpath, buf, path_len))
	    strcpy(buf, "");

	/* read xml data and fill finfo */
	if(strlen(buf)) {
	    dav_fill_finfo(one_res, dc, buf, &finfo);
	}
	
	/* send it to socket */
	if(sock_fullwrite(d->ns, (void*)&finfo, sizeof(finfo))<0) {
	    perror("fatter write");
	    break;
	}
    }
    
#ifdef PROP_CACHE
    if(depth==0 && ret==200)
	add_prop_cache(d->pc, dc->path, &finfo);
#endif 

out:
    safe_free(oldpath);
    xr_free(&c);
    dav_rec_free(&rec);

    return ret;
}

int client_ls(dav_config *dc, char *arg1, char *arg2) {
	return client_ls_attr(dc, arg1, arg2, 1);
}

int client_attr(dav_config *dc, char *arg1, char *arg2) {
	return client_ls_attr(dc, arg1, arg2, 0);
}

int client_setattr(dav_config *dc, char *arg1, char *arg2) {
    int ret=200;
    dav_rec rec;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    struct dav_finfo finfo;

    dav_rec_clear(&rec);
    
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
    
    if(!arg1 || !arg2) goto ret;
    
    /* process arg1 */
    safe_free(dc->path);
    dc->path = reduce_path(arg1);

    if(atoi(arg2)!=sizeof(finfo)) {
	perror("setattr: finfo size is diffrent.");
	goto ret;
    }

    /* read attr data */
    sock_fullread(d->ns, (char*)&finfo, sizeof(finfo));

//#define SETATTR
#ifdef SETATTR
    /* we only care about the time */
    if(finfo.f_mtime) {
	/* make xml */
	sprintf(buf, "\
<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n\
<D:propertyupdate xmlns:D=\"DAV:\">\n\
<D:set>\n\
<D:prop>\n\
<D:davfsmtime>%d</D:davfsmtime>\n\
</D:prop>\n\
</D:set>\n\
</D:propertyupdate>\n", 
		(int)finfo.f_mtime);

	/* set values */
	rec.xml = safe_strdup(buf);
	rec.o.length = strlen(rec.xml);

	/* set attr */
	ret = dav_proppatch(dc, &rec);

#ifdef PROP_CACHE
	invalidate_prop_cache(d->pc, dc->path);
#endif

    }
#endif
#ifdef DEBUG
    printf("finfo->f_size: %ld\n", finfo.f_size);
    printf("finfo->f_isdir: %d\n", finfo.f_isdir);
    printf("finfo->f_atime: %d\n", (int)finfo.f_atime);
    printf("finfo->f_mtime: %d\n", (int)finfo.f_mtime);
    printf("finfo->f_ctime: %d\n", (int)finfo.f_ctime);
#endif

    /* make head */
    dav_mk_res(buf, ret, 0);
    sock_sendline(d->ns, buf);
    
 ret:
    dav_rec_free(&rec);
    
    return ret;
}

int client_quit(dav_config *dc, char *arg1, char *arg2) {
    int ret=200;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
        
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
    
    /* make head */
    dav_mk_res(buf, ret, 0);
    sock_sendline(d->ns, buf);

 ret:    
    /* to quit */
    return -1;
}

int client_statfs(dav_config *dc, char *arg1, char *arg2) {
    int ret=200;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    struct dav_statfs ds;
    
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
    
    /* TODO ??
    ** 
    ** Show we get the information from DAV server?
    **
    **/
    /* Also fake information */
    {
	ds.f_blocks = DEFAULT_BLOCKS;
	ds.f_bfree = DEFAULT_BFREE;
    }
    
    /* make head */
    dav_mk_res(buf, ret, sizeof(ds));
    sock_sendline(d->ns, buf);

    /* send it to socket */
    if(sock_fullwrite(d->ns, (void*)&ds, sizeof(ds))<0)
	perror("fatter write");

 ret:    
    /* to quit */
    return ret;
}

int client_mv(dav_config *dc, char *arg1, char *arg2) {
    dav_rec rec;
    char *src;
    char *des;
    char *ptr=NULL;
    char buf[BUF_SIZE];
    int ret;
    struct descriptor_data *d;
    
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
    
    src = arg1;
    des = arg2;
    
    /* No argument */
    if(!arg1 && !arg2) {
    	dav_mk_res(buf, 404, 0);
	sock_sendline(d->ns, buf);
	goto ret;
    }

    /* get Location to avoid bad gateway */
    safe_free(dc->path);
    dc->path = safe_strdup(des); 
      
    /* delete destination */
    dav_rec_clear(&rec);
    dav_delete(dc, &rec);
    dav_rec_free(&rec);

    /* run put to get Location command */
    dav_rec_clear(&rec);
    dav_put(dc, &rec);

    if(rec.hh.location) 
	ptr = safe_strdup(rec.hh.location);
    dav_rec_free(&rec);
    
    /* delete fake file */
    dav_rec_clear(&rec);
    dav_delete(dc, &rec);
    dav_rec_free(&rec);
    
    dav_rec_clear(&rec);
    if(!ptr) {
	sprintf(buf, "http://%s:%d%s", dc->host, dc->port, des);
	safe_free(ptr);
	rec.dest = safe_strdup(buf);
	printf("Destination: %s\n", rec.dest);
    } else {
	rec.dest = ptr;
    }

#ifdef PROP_CACHE
    invalidate_prop_cache(d->pc, dc->path);
#endif    
    
#ifdef DEBUG
    printf("Destination: %s\n", rec.dest);
#endif

    /* assign old path to new base path */
    safe_free(dc->path);
    dc->path = safe_strdup(src); 

    rec.o.overwrite = 1;   /* set overwrite */
    ret = dav_move(dc, &rec);
    
    /* make head */
    dav_mk_res(buf, ret, 0);
    sock_sendline(d->ns, buf);

 ret:

    /* free recoreds */
    dav_rec_free(&rec);

    return ret;
}

int client_read(dav_config *dc, char *arg1, char *arg2) {
    char *data;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    long offset=0, length=0;
    int ret = 200, fp;

    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
        
    /* no arguments */
    if(!arg1 || !arg2) {
	dav_mk_res(buf, 404, 0);
	sock_sendline(d->ns, buf);
	goto ret;
    }
    
    /* open file */
    sprintf(buf, "%s/dav_cache%s", cache_dir, arg1);

    if(get_hash(atoi(arg1), &fp)!=1) {
    	perror(buf);
	dav_mk_res(buf, 404, 0);
	sock_sendline(d->ns, buf);
	goto ret;
    }

    /* process arg2 */
    data = strtok(arg2, ":\n\r");
    if(data) offset = atoi(data);

    data = strtok(NULL, ":\n\r");
    if(data) length = atoi(data);

    /* no length to read */
    if(!length) {
	dav_mk_res(buf, 404, 0);
	sock_sendline(d->ns, buf);
	goto ret;
    }

    /* set d values */
    d->offset = offset;
    d->length = length;
    d->data = malloc(length);
    d->data_len = 0;

    /* no memory */
    if(!d->data) {
	dav_mk_res(buf, 505, 0);
	sock_sendline(d->ns, buf);
	goto ret;
    }
    
    if(lseek(fp, offset, SEEK_SET) != -1) 
	d->data_len = read(fp, d->data, d->length);
    else perror("client_read");

    if(d->data_len<0) { 
	perror("no_data");
	d->data_len = 0;
	ret = 404;
    }
    
    /* get ralsize from d->data_len */
    dav_mk_res(buf, ret, d->data_len);
    sock_sendline(d->ns, buf);
     
    /* send data to socket */
    if(sock_fullwrite(d->ns, (char*)d->data, d->data_len)<0) {
	perror("fatter write");
    }
    
    /* free d->data */
    safe_free(d->data);
 ret:
    return ret;
}

/*
** write null char to offset, if file is small
**
*/
int dav_offset_write(int fp, long offset, char *data, long data_len) {
    int end;
    int ret = 0;
    int diff, to_write;
    char buf[BUF_SIZE];

    /* let's go end of the file */
    end = lseek(fp, 0, SEEK_END);
    
    /* Oppps, we need to write some garbages */
    if(offset>end) {
	memset(buf, 0x00, sizeof(buf));
	
	for(diff = offset-end; diff>0; diff-=to_write) {
	    to_write = min(BUF_SIZE, diff);
	    to_write = write(fp, buf, to_write);
	    if(to_write<0) 
		break;
	}
    }
    
    if(lseek(fp, offset, SEEK_SET)!=-1) 
	ret = write(fp, data, data_len);
    else perror ("dav_offset_write");
    
    return ret;
}

int client_write(dav_config *dc, char *arg1, char *arg2) {
    char *data;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    long offset=0, length=0;
    int ret = 200, fp;
    
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
    
    /* no arguments */
    if(!arg1 || !arg2) {
	dav_mk_res(buf, 404, 0);
	sock_sendline(d->ns, buf);
	goto ret;
    }
    /* process arg2 */
    data = strtok(arg2, ":\n\r");
    if(data) offset = atoi(data);
    
    data = strtok(NULL, ":\n\r");
    if(data) length = atoi(data);

    /* set d values */
    d->offset = offset;
    d->length = length;
    /* it can be zero */
    d->data = malloc(length+1);

    if(!d->data) {
	dav_mk_res(buf, 505, 0);
	sock_sendline(d->ns, buf);
	goto ret;
    }
    
    /* read data from kernel */
    d->data_len = sock_fullread(d->ns, d->data, length);
    if(d->data_len<0) {
	/* data error */
	dav_mk_res(buf, 403, 0);
	sock_sendline(d->ns, buf);
	goto ret;
    }

    d->data_len = length;
 
    
    /* open file */
    if(get_hash(atoi(arg1),  &fp)!=1) {
	fprintf(stderr, "get_hash:");
	perror(arg1);
	ret = 404;
    } else {
       dav_offset_write(fp, offset, d->data, d->data_len);
    }

    /* result return */
    dav_mk_res(buf, ret, 0);
    sock_sendline(d->ns, buf);
    
    /* free d->data */
    safe_free(d->data);
 ret:
    return ret;
}


int client_unlink(dav_config *dc, char *arg1, char *arg2) {
    int ret;
    dav_rec rec;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
     
    dav_rec_clear(&rec);
   
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
        
    if(!arg1) goto ret;

    /* process arg1 */
    safe_free(dc->path);
    dc->path = reduce_path(arg1);

    fflush(stdout);
    /* delete file */
    ret = dav_delete(dc, &rec);
    fflush(stdout);

    /* make head */
    dav_mk_res(buf, ret, 0);
    sock_sendline(d->ns, buf);

 ret:
#ifdef PROP_CACHE
    invalidate_prop_cache(d->pc, dc->path);
#endif
    dav_rec_free(&rec);
    
    return ret;
}

int do_close(dav_config *dc, dav_rec *rec) {
    char buf[BUF_SIZE];
    int n, tn=0;

    while((n=read(rec->u.in, buf, BUF_SIZE))>0) {
	if(sock_fullwrite(dc->ns, buf, n)<0) break;
	tn+=n;
    }
    
    return tn;
}

int client_close(dav_config *dc, char *arg1, char *arg2) {
    int fp;
    int wish=0;
    int ret=200;
    dav_rec rec;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    
    dav_rec_clear(&rec);
    
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;

    if(!arg1) 
	goto ret;
    
    /* try to put */
    if(del_hash(atoi(arg1), &wish, &fp, buf)==1) {
	if(wish)  {
	    /* process arg1 */
	    safe_free(dc->path);
	    dc->path = reduce_path(buf);
	    
	    /* set rec */
	    rec.send_body  = do_close;
	    rec.u.in = fp;
	    rec.o.length = lseek(fp, 0 , SEEK_END);
	    
	    lseek(fp, 0, SEEK_SET);
	    
#ifdef LOCK
	    /* lock resource */
	    if(client_lock(dc, buf)!=0) {
		rec.lock_token = safe_strdup(buf);
	    }
#endif

	    /* read data from http and save into a  file */
	    ret = dav_put(dc, &rec);
	    
#ifdef LOCK
	    /* unlock resource */
	    client_unlock(dc, rec.lock_token);
#endif
	    
#ifdef PROP_CACHE
	    /* only put */
	    invalidate_prop_cache(d->pc, dc->path);
#endif
	}
	close(fp);
    } else {
	fprintf(stderr, "del_hash: ");
	perror(arg1);
    }
    
    /* delete file */
    sprintf(buf, "%s/dav_cache%s", cache_dir, arg1);
    unlink(buf);
    
 ret:
    /* make head */
    dav_mk_res(buf, ret, 0);
    sock_sendline(d->ns, buf);
    dav_rec_free(&rec);
    
    return ret;
}

/*
** open file and change filename to inode
** return inode to *ino
*/
int dav_inodeopen(int flags, mode_t mode, long *ino) {
    struct stat sbuf;
    struct timeb tbuf;
    char old[BUF_SIZE];
    char new[BUF_SIZE];
    int ret;

    ftime(&tbuf);
    sprintf(old, "%s/dav_cache%ld.%d", cache_dir, (long)tbuf.time, tbuf.millitm);
    
    umask(022);
    ret = creat(old, mode);
    if(ret==-1)
	return ret;

    fstat(ret, &sbuf);
    close(ret);

    /* make new filename */
    sprintf(new, "%s/dav_cache%ld", cache_dir, (long)sbuf.st_ino);
    rename(old, new);

    *ino = (long)sbuf.st_ino;
    return open(new, flags, mode);
}

int client_open(dav_config *dc, char *arg1, char *arg2) {
    int ret=200;
    dav_rec rec;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    long fileid;
    int fp, wish;

    dav_rec_clear(&rec);

    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
        
    if(!arg1 || !arg2) {
	ret = 403;
	goto ret;
    }

    wish = atoi(arg2);
    
    /* inode file open */
    if((fp=dav_inodeopen(O_RDWR|O_CREAT|O_TRUNC, 0666, &fileid))==-1) {
	perror("dav_inodeopen");
	ret = 505;
	goto ret;
    }
    
    /* add to hash */
    add_hash(fileid, atoi(arg2), fp, arg1);

    /* read mode */
   // if(!wish) {
    if(1) {
	/* set rec */
	rec.u.out = fp;
	
	/* process arg1 */
	safe_free(dc->path);
	dc->path = reduce_path(arg1);
	
	/* read data fro http and save into a  file */
	ret = dav_get(dc, &rec);
    }
 ret:
    dav_mk_res(buf, ret, fileid);
    sock_sendline(d->ns, buf);

#ifdef SHOW_CMD
    log(buf);
#endif

    dav_rec_free(&rec);
    return ret;
}


int client_create(dav_config *dc, char *arg1, char *arg2) {
    int ret=200;
    dav_rec rec;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    long fileid;
    int fp, wish;
    
    dav_rec_clear(&rec);
    
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
    
    if(!arg1) { 
	ret = 403;
	goto ret;
    }

    wish = 1;
    
    /* inode file open */
    if((fp=dav_inodeopen(O_RDWR|O_CREAT|O_TRUNC, 0666, &fileid))==-1) {
	perror("dav_inodeopen");
	ret = 505;
	goto ret;
    }
        
    /* add to hash */
    add_hash(fileid, wish, fp, arg1);
    
    /* make zero file */
    rec.o.length=0;
    
    /* process arg1 */
    safe_free(dc->path);
    dc->path = reduce_path(arg1);
    
    ret = dav_put(dc, &rec);
 ret:
    dav_mk_res(buf, ret, fileid);
    sock_sendline(d->ns, buf);

#ifdef SHOW_CMD
    log(buf);
#endif

    dav_rec_free(&rec);
    return ret;
}


int client_mkdir(dav_config *dc, char *arg1, char *arg2) {
    int ret;
    dav_rec rec;
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    
    dav_rec_clear(&rec);
  
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
        
    /* process arg1 */
    safe_free(dc->path);
    dc->path = reduce_path(arg1);

    /* delete file */
    ret = dav_mkcol(dc, &rec);

    /* make head */
    dav_mk_res(buf, ret, 0);
    sock_sendline(d->ns, buf);
    
 ret:
    dav_rec_free(&rec);
    
    return ret;
}


int client_rmdir(dav_config *dc, char *arg1, char *arg2) {
    return client_unlink(dc, arg1, arg2);
}


/* no command */
int client_badcmd(dav_config *dc, char *arg1, char *arg2) {
    char buf[BUF_SIZE];
    struct descriptor_data *d;
    
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
        
    /* make head */
    dav_mk_res(buf, 403, 0);
    sock_sendline(d->ns, buf);

ret:    
    return 0;
}

int client_trunc(dav_config *dc, char *arg1, char *arg2) {
    int fp, fp_old; 
    int ret=403;
    int n, to_copy, to_read, all_read=0;
    char buf[BUF_SIZE], buf2[BUF_SIZE];
    struct descriptor_data *d;
    
    d = (struct descriptor_data *)dc->userdata;
    if(!d) goto ret;
    
    if(!arg1 || !arg2) 
	goto ret;
    
    /* trunc size */
    to_copy = atoi(arg2);

    /* get fp and delete */
    if(get_hash(atoi(arg1), &fp)==1) {
	close(fp);
	
	sprintf(buf, "%s/dav_cache%s", cache_dir, arg1);
	/* if we have a trunc size */
	if(to_copy) {
	    sprintf(buf2, "%s/dav_cache%s-back", cache_dir, arg1);
	    rename(buf, buf2);
	}
	
	fp = open(buf, O_RDWR|O_CREAT|O_TRUNC, 0666);
	if(fp==-1) {
	    perror("dav_inodeopen");
	    ret = 505;
	    goto ret;
    	}

	/* if we have a second arg on trunc */
	if(to_copy) {
	    if((fp_old = open(buf2, O_RDONLY))!=0) {
		while(to_copy>all_read) {
		    to_read = to_copy-all_read>BUF_SIZE?
		      BUF_SIZE:to_copy-all_read;
		    n = read(fp_old, buf, to_read);
		    if(n<=0) break;
		    write(fp, buf, n);
		    all_read+=n;

		}
		close(fp_old);
	    }
	    remove(buf2);
	}
	
	/* add new hash list */
	if(set_hash(atoi(arg1), fp)!=1) 
	    goto ret;
	
	ret = 200;
    }
       
ret:
    /* make head */
    dav_mk_res(buf, ret, 0);
    sock_sendline(d->ns, buf);
    
    return ret;
}













