/* 
** davfsd main program 
** Copyright (C) 2001 SungHun Kim <hunkim@kebi.com>
**
** This library and program is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**   
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
** MA 02111-1307, USA
*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <davfsd.h>
#include <libdav.h>
#include <nsocket.h>

int CloseDavfsd;
int MotherDesc;

void child_handle( int sig_no )
{
    waitpid( 0, NULL, WNOHANG );
    signal( SIGCHLD, child_handle);
}

int save_pid(void) {
    FILE *fp;

    if(access(PID_FILE, R_OK)!=-1)
	fprintf(stderr, "A pid file already exists!\n");

    if((fp=fopen(PID_FILE, "w"))==NULL) {
	perror(PID_FILE);
	return 0;
    }
    
    fprintf(fp, "%d", getpid());
    fclose(fp);

    return 1;
}

    
/* from samba 2.0 mount */
static void daemonize(void)
{
    int  status;
    pid_t child_pid;
    
    if ((child_pid = fork()) < 0) {
	log("could not fork\n");
    }
    
    if (child_pid > 0) {
#ifdef DAV_WAIT
        int j
	while( 1 ) {
	    j = waitpid( child_pid, &status, 0 );
	    if( j < 0 ) {
		if( EINTR == errno ) {
		    continue;
		}
		status = errno;
	    }
	    break;
	}
#endif
	/* If we get here - the child exited with some error status */
	exit(status);
    }
    
    chdir("/");
}
 

void nonblock(socket_t s)
{
    int flags;
    
    flags = fcntl(s, F_GETFL, 0);
    flags |= O_NONBLOCK;
    if (fcntl(s, F_SETFL, flags) < 0) {
	perror("SYSERR: Fatal error executing nonblock (comm.c)");
	exit(1);
    }
}

/*
 * get_bind_addr: Return a struct in_addr that should be used in our
 * call to bind().  If the user has specified a desired binding
 * address, we try to bind to it; otherwise, we bind to INADDR_ANY.
 * Note that inet_aton() is preferred over inet_addr() so we use it if
 * we can.  If neither is available, we always bind to INADDR_ANY.
 */

struct in_addr *get_bind_addr()
{
    static struct in_addr bind_addr;
    
    /* Clear the structure */
    memset((char *) &bind_addr, 0, sizeof(bind_addr));
    
    /* Put the address that we've finally decided on into the logs */
#ifdef SHOW_CMD
    if (bind_addr.s_addr == htonl(INADDR_ANY))
	log("Binding to all IP interfaces on this host.");
    else
	log("Binding only to IP address %s", inet_ntoa(bind_addr));
#endif
    
    return (&bind_addr);
}


void close_socket(struct descriptor_data *d) 
{
    close(d->descriptor);
    safe_free(d->ns);

    safe_free(d->output);
    safe_free(d->data);
    safe_free(d->orig_path);
    
    /* safe dc in d */
    dav_dc_free(&(d->dc));

    /* safe free itself */
    safe_free(d);
}

/*
 * init_socket sets up the mother descriptor - creates the socket, sets
 * its options up, binds it, and listens.
 */
socket_t init_socket(int port)
{
    socket_t s;
    struct sockaddr_in sa;
    int sockopton = 1;

    if ((s = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
	perror("SYSERR: Error creating socket");
	exit(1);
    }
    
    /* added by Anthony Baxter
    ** Thu Aug  9 16:02:00 2001
    */
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, 
		   (char *)&sockopton, sizeof(int) ) == -1 ) {
	perror("SYSERR: setsockopt(SO_REUSEADDR)");
	close(s);
	exit(1);
    }

    /* Clear the structure */
    memset((char *)&sa, 0, sizeof(sa));
    
    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);
    sa.sin_addr = *(get_bind_addr());
    
    if (bind(s, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
	perror("SYSERR: bind");
	close(s);
	exit(1);
    }
    //nonblock(s);
    listen(s, 5);
    
    return (s);
}

void log(const char *format, ...)
{
  FILE *logfile = stderr;
  va_list args;
  static long count=0;

  time_t ct = time(0);
  char *time_s = (char*)asctime(localtime(&ct));
  
  if (format == NULL)
      format = "SYSERR: log() received a NULL format.";
  
  time_s[strlen(time_s) - 1] = '\0';
  
  fprintf(logfile, "[%ld] %-15.15s :: ", count++, time_s + 4);
  
  va_start(args, format);
  vfprintf(logfile, format, args);
  va_end(args);
  
  if(!strchr(format, '\n'))
      fprintf(logfile, "\n");
  fflush(logfile);
}


/*
** parse server information and command 
** 
** fill output, outdata, data_len
**
** return : 0 ; nothing
** return : -1; disconnect 
*/
int do_command(char *command, struct descriptor_data *d) {
    int ret;
    char req[BUF_SIZE];
    char *com;
    char *delm = "\t\n\r";
    char *arg1=NULL, *arg2=NULL;
    struct dav_finfo finfo;
    dav_config *dc = &(d->dc);
    char *orig_path = dc->path;
    
    int (*client_func)(dav_config *dc, char *arg1, char *arg2) = NULL;
    
#ifdef SHOW_CMD
    log(command);
#endif

    com = strtok(command, delm);
    if(!com) {
	perror("No command");
	return 0;
    }
    
    arg1 = strtok(NULL, delm);
    arg2 = strtok(NULL, delm);

    switch(*com) {
	
	/*
	** ls
	**
	** 
	*/
    case 'l':    
	if(!strcmp(com, "ls"))
	    client_func = client_ls;
	
	break;
	
	/*
	** attr : get attribute
	**
	** attr <file|dir>
	*/
    case 'a':
	if(!strcmp(com, "attr")) {
	    char *path_arg1=NULL;
#ifdef PROP_CACHE
	    char *path=NULL;

	    /* check cache */
	    if(arg1) 
		path_arg1 = strdup(arg1);

	    if(path_arg1 && (path=reduce_path(path_arg1)) && 
	       get_prop_cache(d->pc, path, &finfo)==1) {
#ifdef DEBUG
		printf("PROPFIND [%s] from cache\n", path);
#endif
		return client_attr_cache(&(d->dc), &finfo);
#endif
	    } else {
		client_func = client_attr;
	    }
	    if(path_arg1)
		safe_free(path_arg1);
	
	}
	break;
	
	/*
	** read : read from file
	**
	** read <file> <offset:size>
	*/

    case 'r':
	if(!strcmp(com, "read")) 
	    client_func = client_read;
	else if(!strcmp(com, "rmdir"))
	    client_func = client_rmdir;
	break;

	/*
	** write : write to file
	**
	** write <file> <offset:size>
	**
	*/
    case 'w':
	if(!strcmp(com, "write"))
	    client_func = client_write;
	break;
	
	/*
	** unlink
	** unlink <file>
	*/	
    case 'u':
	if(!strcmp(com, "unlink"))
	    client_func = client_unlink;
	break;
	
    case 'm':
	if(!strcmp(com, "mkdir"))
	    client_func = client_mkdir;
	else if(!strcmp(com, "mv"))
	    client_func = client_mv;
	break;
	
	/*
	** open <filename>
	*/
    case 'o':
	if(!strcmp(com, "open"))
	    client_func = client_open;
	break;
	
	/*
	** close <fid>
	*/
    case 'c':
	if(!strcmp(com, "close"))
	    client_func = client_close;
	else if(!strcmp(com, "create"))
	    client_func = client_create;
	break;
	

	/*
	** trunc <fid> <length>
	*/
    case 't':
	if(!strcmp(com, "trunc"))
	    client_func = client_trunc;
	break;

	/*
	** quit 
	*/
    case 'q':
	/* this function should result -1 */
	if(!strcmp(com, "quit"))
	    client_func = client_quit;
	
	break;

	/*
	** setattr 
	*/
    case 's':
	if(!strcmp(com, "setattr"))
	    client_func = client_setattr;
	else if(!strcmp(com, "statfs"))
	    client_func = client_statfs;
	break;

    default:
	puts("Command not found!!");
	client_func = client_badcmd;
	break;

    }

    /* set connected flag */
    if(dav_reopen(dc)==NULL) {
	perror(dc->host);
	dc->connected = 0;
	
	/* send error message to kernel */
	dav_mk_res(req, 500, 0);
	sock_sendline(d->ns, req);
	
	return 0;
    } else {
	dc->connected = 1;
    }

    
    /*
    ** do client_func, if available
    */
    if(client_func) 
	ret = client_func(dc, arg1, arg2);
    else 
	ret = client_badcmd(dc, arg1, arg2);

    dav_only_close(dc);
    
    /* reassing path, because path can cahge after method call */
    if(dc->path!=orig_path) {
	safe_free(dc->path);
	dc->path = safe_strdup(d->orig_path);
    }
    return ret;
}

/* create new descriptor */
struct descriptor_data *new_descriptor(socket_t desc) 
{
    char buf[BUF_SIZE];
    struct descriptor_data *newd; 
    char *url=NULL, *proxy=NULL, *id=NULL, *pass=NULL;
    dav_config *dc;
    
    /* make */
    newd = (struct descriptor_data *)malloc(sizeof(struct descriptor_data));
    if(newd == NULL) {
	perror("ndew");
	return NULL;
    }
    memset(newd, 0x00, sizeof(struct descriptor_data));
   
    /* Assign temprary dc pointer */
    dc = &(newd->dc);

    newd->descriptor = desc;
    newd->ns = create_sock(desc);
    if(!newd->ns) {
	safe_free(newd);
	perror("make ns");
	return NULL;
    }

    /* Grreting and get some information */
    sprintf(buf, "%s\t%s\tI need URL<CR><LF>proxy<CR><LF>ID<CR><LF>", PACKAGE, VERSION);
    sock_sendline(newd->ns, buf);
    
    /* read url */
    if(sock_readline(newd->ns, buf, BUF_SIZE-1) >=0 && *buf!='\n' && *buf!='\r' )
	url = safe_strdup(strtok(buf, "\n\r"));
    
    /* read proxy */
    if(sock_readline(newd->ns, buf, BUF_SIZE-1) >= 0 &&  *buf!='\n' && *buf!='\r' )
	proxy = safe_strdup(strtok(buf,"\n\r"));
    	  
    /* read id */
    if(sock_readline(newd->ns, buf, BUF_SIZE-1) >= 0 && *buf!='\n' && *buf!='\r' )
	id = safe_strdup(strtok(buf,"\n\r"));
    
    /* read pass */
    if(sock_readline(newd->ns, buf, BUF_SIZE-1) >= 0 && *buf!='\n' && *buf!='\r' )
	pass = safe_strdup(strtok(buf, "\n\r"));

#ifdef DEBUG
    printf("----------------------------------------\n");
    printf("First communication : %s %s %s %s\n",
	url?url:"NULL",
	proxy?proxy:"NULL",
	id?id:"NULL",
	pass?pass:"NULL");
    printf("----------------------------------------\n");
#endif

    /* make dc */

    /* clear dc structure */
    memset(dc, 0x00, sizeof(dav_config));
    
    /* assign d */
    dc->userdata = (void*)newd;

    if(dav_prepare(dc, url, id, pass, proxy)==-1) {
	perror("dav_prepare");
	return NULL;
    }
    
    newd->orig_path = safe_strdup(dc->path);
    
    safe_free(url);
    safe_free(id);
    safe_free(pass);
    safe_free(proxy);

    return newd;
}

/* process input data */
int process_input(struct descriptor_data *d) {
    char command[BUF_SIZE];
    int ret;

    memset(command, 0x00, sizeof(command));
    
    /* read command line */
re1:
    ret = sock_readline(d->ns, command, BUF_SIZE-1);
    if(ret==-2) { /* time out */
#ifdef SHOW_CMD
	fprintf(stderr, "waitting for input\n");
#endif
	goto re1;
    }
    if(ret<=0) 
	return ret;
    
    return do_command(command, d);
}

int process_output(struct descriptor_data *d) {
    
    if(d->output) {
	write(d->descriptor, d->output, strlen(d->output));
	safe_free(d->output);
    }
    return 0;
}


int davfsd(socket_t mother_desc) {
    struct descriptor_data *d;
    struct sockaddr_in cliaddr;
    int pid, connfd;
    int  len = sizeof(cliaddr);
    
    while(!CloseDavfsd) {
	if( (connfd = accept(mother_desc, (struct sockaddr*)&cliaddr, &len)) < 0){
	    if( errno == EINTR ){
		continue;
	    }else if(errno == ETIMEDOUT || errno == ECONNRESET){
		continue;
	    }else{
		printf("[errno:%d] Accept Error:%s", errno, strerror(errno));
		exit(0);
	    }
	}

	/* fork */
#ifndef GDB
	if( (pid = fork() ) == 0 ){
#else
	if(1){
#endif
	    close(mother_desc);
	    
	    /* Host Check
	    inet_ntop(AF_INET, &cliaddr.sin_addr, buff, sizeof(buff));
	    
	    if( !auth_trusthost(buff) ){
		close(connfd);
		auth_exit(1);
	    }
	    */

	    /* make new d*/
	    d = new_descriptor(connfd);
	    if(d == NULL) {
		perror("fail to make new_descriptor");
		exit(0);
	    }
	 
#ifdef PROP_CACHE
	    /* init prop cache */
	    init_prop_cache(d->pc);
	    memset(d->pc, 0x00, sizeof(d->pc));
#endif
	    /* 0: quit, -1: error 1: continue*/
	    while( process_input(d) >= 0);
	    close_socket(d);
#ifdef SHOW_CMD
	    log("Close a client");
#endif

#ifdef PROP_CACHE
	    /* free prop cache */
	    free_prop_cache(d->pc);
#endif
	    
	    exit(0);
	} else {  
	    close(connfd);
	}
    }
    close(mother_desc);

    return 0;
}
 
void close_all(int sig) {
    CloseDavfsd = 1;
    close(MotherDesc);

    /* umount all davfs */
    umount_all();

    /* delete pid file */
    remove(PID_FILE);

    log("close_all");
    exit(0);
}

int usage(void) {
    return 0;
}

char *cache_dir = CACHE_DIR;
int main(int argc, char *argv[]) {
    socket_t mother_desc;
    int o;

    /* sock initialize */
    sock_init();

    /* get option */
    while((o=getopt(argc, argv, "d:h:k"))!=-1) {
   	switch(o) {
	case 0 :
	case 'h' :
	    usage();
	    return 0;
	case 'd':
	    cache_dir = optarg;
	    break;
	    
	case 'k':
	    printf("K: %s\n", optarg);
	    break;
	    
	}
    }    

    /* check CACHE dir */
    if(access(cache_dir, W_OK)==-1) 
	if(mkdir(cache_dir, 04755)==-1) {
	    perror(cache_dir);
	    exit(-1);
	}
	
    signal(SIGINT, close_all);
    signal(SIGILL, close_all);
    signal(SIGBUS, close_all);
    signal(SIGKILL, close_all);
    signal(SIGSEGV, close_all);
    signal(SIGTERM, close_all);
    signal( SIGCHLD, child_handle);

    /* fork child */
#ifndef GDB
    daemonize(); 
#endif

    /* save pid */
    save_pid();

    MotherDesc = mother_desc = init_socket(PORT);

    davfsd(mother_desc);
    
    shutdown(mother_desc, 0);

    log("Closing mother socket.");
    close(mother_desc);

    exit(0);
}








