/* 
** Hashed linked list for PROPFIND
** Copyright (C) 2001 SungHun Kim <hunkim@kebi.com>
**
** This library and program is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**   
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
** MA 02111-1307, USA
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/timeb.h>

#include <davfsd.h>
#include <propcache.h>

/* hash function */
static int prop_hash(const char *str) {
    int len = (strlen(str)+1) >> 1;  /* Number of shorts */
    unsigned short *ptr = (unsigned short *)str;
    unsigned hash = len;
    while (--len >= 0) {
	hash = (hash << 1) ^ *ptr++;
    }
    return (hash%PROP_CACHE_SIZE);
}


/* initialize */
int init_prop_cache(prop_cache **pc) {
    memset(pc, 0x00, sizeof(prop_cache*)*PROP_CACHE_SIZE);
    return 1;
}

/* free prop cache memory */
int free_prop_cache(prop_cache **pc) {
    int i;
    int ret = 0;

    for(i=0;i<PROP_CACHE_SIZE;i++) {
	if(pc[i]!=NULL) {
	    safe_free(pc[i]->finfo);
	    safe_free(pc[i]->path);
	    safe_free(pc[i]);
	    ret++;
	}
    }

    return ret;
}

/* add cache */
int add_prop_cache(prop_cache **pc, char *path, struct dav_finfo *finfo) {
    int hash_no = prop_hash(path);
    
#ifdef DEBUG
    printf("Add %s at %d bucket\n", path, hash_no);
#endif

    if(pc[hash_no]) 
	safe_free(pc[hash_no]->finfo), safe_free(pc[hash_no]->path);
    else
	pc[hash_no] = malloc(sizeof(prop_cache));
    
    if(!pc[hash_no]) {
	perror("malloc pc_hash_no");
	return -1;
    }

    memset(pc[hash_no], 0x00, sizeof(prop_cache));
    pc[hash_no]->finfo = malloc(sizeof(struct dav_finfo));
    pc[hash_no]->path = safe_strdup(path);
        
    if(!(pc[hash_no]->finfo) || !(pc[hash_no]->path)) {
	perror("malloc pc finfo");
	safe_free(pc[hash_no]);
	return -1;
    }
    
    memcpy(pc[hash_no]->finfo, finfo, sizeof(struct dav_finfo));
    pc[hash_no]->ctime = time(NULL);
    
    return hash_no;
}


/* get cache
**
** return value 
**   has cache 1
**   no cache  0
**   error -1 
*/

int get_prop_cache(prop_cache **pc, char *path, struct dav_finfo *finfo) {
    int hash_no = prop_hash(path);
#ifdef DEBUG
	printf("S_GET:%s:%d\n", path, hash_no);
#endif

    if(!pc[hash_no] || !pc[hash_no]->finfo || pc[hash_no]->isdirty ||  
       time(NULL) - pc[hash_no]->ctime > CACHE_TIMEOUT ||
       !pc[hash_no]->path || strcmp(pc[hash_no]->path, path)) {
	return 0;
    }
    
    memcpy(finfo, pc[hash_no]->finfo, sizeof(struct dav_finfo));
#ifdef DEBUG
	puts("E_GET");
#endif
    return 1;
}

int invalidate_prop_cache(prop_cache **pc, char *path) {
    int hash_no = prop_hash(path);
    
    if(pc[hash_no] && pc[hash_no]->finfo && !pc[hash_no]->isdirty && 
       pc[hash_no]->path &&strcmp(pc[hash_no]->finfo->f_name, path)) {
	pc[hash_no]->isdirty = 1;
	return 1;
    }
    return 0;
}
