/* 
** Dav library head file
** Copyright (C) 2001 SungHun Kim <hunkim@kebi.com>
**
** This library and program is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**   
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
** MA 02111-1307, USA
*/

#ifndef __LIBDAV_H__
#define __LIBDAV_H__

#include <nsocket.h>
#include <xmlparse.h>

/* seperate XML */
#define SEP_XML 2

/* general buff size */
#define BUF_SIZE 1024

/* default lock time out */
#define LOCK_TIMEOUT "Second-60"

/* default ports */
#define DEFAULT_PORT 80
#define DEFAULT_SSL_PORT 443

/* handler for dav connection */
typedef struct {
    int sock;                /* socket */
    nsocket *ns;             /* nsocket */
    char *host;              /* host name */
    char *scheme;            /* protocol scheme */
    char *path;              /* path */
    char *errment;           /* error message */
    char *id;                /* user id */
    char *pass;              /* user password */
    char *auth_type;         /* auth type (Basic) */
    char *authorization;     /* autorization header */
    int port;                /* server port */
    int errorno;             /* error number */
    int connected;           /* conneected information */
    
    char *proxy_scheme;            /* proxy protocol scheme */
    char *proxy_host;        /* proxy host name */
    int proxy_port;          /* proxy port number */
    
    void *userdata;          /* user data pointer */
} dav_config;


/* header information of response */
typedef struct {
    int ret_code;            /* return code */   
    int content_length;      /* contend length */
    int has_content_length;  /* contend length */
    int chunked;             /* is chunked data? */

    /* need to be free after parse head */
    char *allow;             /* allow string */
    char *content_type;      /* content type string */
    char *location;          /* location string */
    char *lock_token;        /* lock token */
} http_head;

/* response record */
typedef struct dav_rec_t {
    int method;             /* request method */
    http_head hh;           /* head information */

    char *dest;             /* destnation string for move, copy */
    
    /* union for values */
    union {
	int in;             /* file handler for put */
	int out;            /* file handler for get */
    } u;

    /* union for options */
    union {
	int depth;          /* depth for propfind */
	int overwrite;      /* options for moving or copying */
	int length;         /* data length to request */
    } o;

    /* process functions */
    int (*start_func)(dav_config *, struct dav_rec_t *); /* do someting before get body */
    int (*do_func)(dav_config *, struct dav_rec_t *);    /* body process function */ 
    int (*end_func)(dav_config *, struct dav_rec_t *);   /* do something after get body */
    int (*send_body)(dav_config *, struct dav_rec_t *);  /* send whole data */
    
    /* you can pass your data. You should free after using it */
    void *userdata;

    /* lock toketn */
    char *lock_token;

    /* out put for xml */
    char *xml;

    /* XML parser */
    XML_Parser xml_p;       
} dav_rec;


enum dav_errno { OK, 
	     NULLFILE,
	     NOSERV};


#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif


#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif

/*
** 
** prototypes of functions
**
*/
void safe_free(void *ptr);
char *safe_strdup(const char *ptr);
void dav_rec_clear(dav_rec *rec);
void dav_rec_free(dav_rec *rec);
int dav_prepare(dav_config *dc, char *url, char *id, char *pass, char *proxy);
nsocket * dav_open(dav_config *dc, char *url, char *id, char *pass, char *proxy);
nsocket * dav_reopen(dav_config *dc);
int dav_only_close(dav_config *dc);
int dav_close(dav_config *dc);

/* method functions */
int dav_put       (dav_config *dc, dav_rec *rec);
int dav_get       (dav_config *dc, dav_rec *rec);
int dav_propfind  (dav_config *dc, dav_rec *rec);
int dav_proppatch (dav_config *dc, dav_rec *rec);
int dav_mkcol     (dav_config *dc, dav_rec *rec);
int dav_delete    (dav_config *dc, dav_rec *rec);
int dav_copy      (dav_config *dc, dav_rec *rec);
int dav_move      (dav_config *dc, dav_rec *rec);
int dav_options   (dav_config *dc, dav_rec *rec);
int dav_head      (dav_config *dc, dav_rec *rec);
int dav_get_head  (dav_config *dc, dav_rec *rec);
int dav_lock      (dav_config *dc, dav_rec *rec);
int dav_unlock    (dav_config *dc, dav_rec *rec);
int dav_dc_free   (dav_config *dc);
int dav_parseurl  (dav_config *dc, char *url);

/* date process */
time_t dav_parse_date(const char *date);
#endif






