/*
 * mount.pvfs.c copyright (c) 1997 Clemson University, all rights reserved.
 *
 * Written by Rob Ross, based somewhat on nfs mount code by Rick Sladkey
 * <jrs@world.std.com>.
 *
 * Modified by Dan Nurmi <nurmi@mcs.anl.gov> to get /etc/mtab right.
 * Option parsing code based somewhat on smb mount code.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.eng.clemson.edu
 * 
 * Modified by Sung Kim  hunkim@kebi.com
 * for davfs file system. Aug 21, 2001
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>

#include <pwd.h>
#include <grp.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/mount.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <mntent.h>

#include <signal.h>
#include <errno.h>
#include <mount.davfs.h>

#include <nsocket.h>
#include <libdav.h>
#include <config.h>

/*
#define AUTOFS
*/

/* source from samba 2.0 */
static void exit_parent(int sig)
{
  /* parent simply exits when child says go... */
  exit(0);
}


/* This is from ncpmount.c       */
/* Returns 0 if the filesystem is in the kernel after this routine
   completes */
static int load_davfs(void)
{
  FILE *ffs;
  char s[1024];
  char *p, *p1;
  pid_t pid;
  int status;

  /* Check if ncpfs is in the kernel */
  ffs = fopen("/proc/filesystems", "r");

  if (ffs == NULL) {
    perror("Error: \"/proc/filesystems\" could not be read:");
    return -1;
  }
  p = NULL;
  while (!feof(ffs)) {
    p1 = fgets(s, sizeof(s), ffs);
    if (p1) {
	p = strstr(s, "davfs");
      if (p) {
	break;
      }
    }
  }
  fclose(ffs);

  if (p) {
    return 0;
  }
  /* system() function without signal handling, from Stevens */

  if ((pid = fork()) < 0) {
    return 1;
  } else if (pid == 0) {
    /* child */
    execl("/sbin/insmod", "insmod", "davfs", NULL);
    _exit(127);			/* execl error */
  } else {
    /* parent */
    while (waitpid(pid, &status, 0) < 0) {
      if (errno != EINTR) {
	status = -1;
	break;
      }
    }
  }
  return status;
}


static void daemonize(void)
{
  int j, status;
  pid_t child_pid;

  signal(SIGTERM, exit_parent);

  if ((child_pid = sys_fork()) < 0) {
    DEBUG2(0, ("could not fork\n"));
  }

  if (child_pid > 0) {
    while (1) {
      j = waitpid(child_pid, &status, 0);
      if (j < 0) {
	if (EINTR == errno) {
	  continue;
	}
	status = errno;
      }
      break;
    }
    /* If we get here - the child exited with some error status */
    exit(status);
  }

  signal(SIGTERM, SIG_DFL);
  chdir("/");
}


int connect_davfsd(char *hostname, int portnum)
{
  int fd;
  struct hostent *hp;
  unsigned long laddr;
  struct in_addr addr;
  struct sockaddr_in sa;

  laddr = (unsigned long) inet_addr(hostname);

  if ((int) laddr == -1) {
    /* inet_addr failed. */
    hp = gethostbyname(hostname);
    if (hp == NULL)
      return -1;
    memcpy(&addr, hp->h_addr, hp->h_length);
  } else {
    addr.s_addr = laddr;
  }


  /* Create the socket */
  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0)
    return fd;

  /* Connect the nsocket */
  sa.sin_family = AF_INET;
  sa.sin_port = htons(portnum);	/* host -> net byte orders */
  sa.sin_addr = addr;
  if (connect(fd, (struct sockaddr *) &sa, sizeof(struct sockaddr_in)) < 0) {
    return -1;
  }
  /* Success - return the nsocket */
  return (fd);
}

int mount_parseurl(char *buf, struct dav_mount_data *in_clargs)
{
    dav_config dc;	
    int ssl_mode = 0;
    
    memset(&dc, 0x00, sizeof(dc));
    if (dav_parseurl(&dc, buf) || dc.host == NULL) {
	perror("parse url");
	exit(-1);
    }
    
    /* is it ssl? */
    if(dc.scheme && !strcmp(dc.scheme, "https")) {
	in_clargs->ssl_mode = ssl_mode = 1;
    } else {
	in_clargs->ssl_mode = ssl_mode = 0;
    }
    
    /* not assigned */
    if(dc.port == -1)
	dc.port = (ssl_mode)?DEFAULT_SSL_PORT:DEFAULT_PORT;  
    
    if(ssl_mode) {
#ifdef ENABLE_SSL    
#else
	perror("Not support SSL!");
	exit(0);
#endif
    }
    /* copy to in_clargs */
    in_clargs->port = dc.port;
    strncpy(in_clargs->host, dc.host, DAV_HOSTLEN);
    strncpy(in_clargs->path, dc.path, DAV_PATHLEN);
    
    dav_dc_free(&dc);
    return 0;
}

#ifdef AUTOFS
/* patched by Stephen Coursen on Aug 27, 2001      */ 
/* Parse commandline arguments passed in via mount */
int parse_mount_args(int argc, char *argv[], struct dav_mount_data *in_clargs, char *url) {
    int opt;
    char buf[DAV_HOSTLEN];
    char options[255];
    char *options_array[255];
    char *temp, *optname, *optvalue;
    int c, array_size;
    extern char *optarg;
    extern int optind, opterr, optopt;
    
    /* mount point */
    if(argv[2]) 
      strcpy(in_clargs->mpoint, argv[2]);
    
    argc -= 2;
    argv += 2;
    
    /* Set default values */
    //strcpy(in_clargs->amode, "rw");
    
    /* u, gid test */
    in_clargs->uid = in_clargs->gid = 500;

    /* Start parsage */
    while ((opt = getopt (argc, argv, "o:")) != EOF)	{
       switch (opt) {
        case 'o':
 	 strncpy( options, optarg, 255 );
 	 break;
        default:
 	 return 1;
       }
    }
    temp = strtok( options, "," );
    c = 0;
    while( temp ) {
       options_array[c] = temp;
       c++;
       temp = strtok( NULL, "," );
    }
    options_array[c] = NULL;
    array_size = c;
    for( c = 0; c < array_size; c++ ) {
       optname = strtok( options_array[c], "=" );
       optvalue = strtok( NULL, "=" );
       if( !optvalue ) {
 	 /* The only single argument option I'm concerned with 
 	  are ro or rw */
 	 if( !strcmp( optname, "ro" ) || !strcmp( optname, "rw" ) )
 	   strcpy( in_clargs->amode, "r" );
       }
       else {
 	 if( !strcmp( optname, "host" ) ) {
 	    strcpy( buf, optvalue );
 	    if(*(buf+strlen(buf)-1)!='/') 
 	      strcat(buf, "/");
 	    strcpy( url, buf );
 	    mount_parseurl( buf, in_clargs );
 	 }
 	 if( !strcmp( optname, "user" ) )
 	   strcpy( in_clargs->id, optvalue );
 	 if( !strcmp( optname, "pass" ) )
 	   strcpy( in_clargs->pass, optvalue );
       }
    }
    dprintf("Parsed Args: url=%s host=%s, path=%s, local_mountpoint=%s, amode=%s, port=%d\n", 
 	   buf, in_clargs->host, in_clargs->path, in_clargs->mpoint, 
 	   in_clargs->amode, in_clargs->port);
    return 0;
 }
#endif

/* ResolveAddr - returns struct in_addr holding address of a host
 *               specified by the character string name, which can
 *               either be a host name or IP address in dot notation.
 */
int ResolveAddr(char *name, struct in_addr *addr)
{
  struct hostent *hep;

  if (!inet_aton(name, addr)) {
    if (!(hep = gethostbyname(name))) {
      fprintf(stderr, "ERROR: cannot resolve remote hostname: %s \n", name);
      return (1);
    } else {
      bcopy(hep->h_addr, (char *) addr, hep->h_length);
      return (0);
    }
  }
  return (0);
}				/* end of ResolveAddr() */


/* 
 * parse_args: function takes argc/argv pair and a pointer to a struct
 * davfs_mount_clargs.  The struct contains variables that need to be
 * filled with information from the commandline to be later used back
 * in main().
 */
int parse_args(int argc, char *argv[], struct dav_mount_data *in_clargs)
{
 
  int opt;
  char buf[DAV_HOSTLEN];
  extern char *optarg;
  struct passwd *passwd;
  struct group *group;

  /* url */
  if (argv[1]) {
      strcpy(buf, argv[1]);
      
      /* add end slash */
      if (*(buf + strlen(buf) - 1) != '/')
	  strcat(buf, "/");
      
      mount_parseurl(buf, in_clargs);
  }
  
  /* mount point */
  if (argv[2])
      strcpy(in_clargs->mpoint, argv[2]);
  
  argc -= 2;
  argv += 2;
  
  /* Set default values */
  //strcpy(in_clargs->amode, "rw");
  strcpy(in_clargs->amode, "r");
  
  /* Start parsage */
  while ((opt = getopt(argc, argv, "u:p:c:U:G:")) != EOF) {
      switch (opt) {
      case 'u':
	  strcpy(in_clargs->id, optarg);
	  break;

      case 'p':
	  strcpy(in_clargs->pass, optarg);
	  break;

      case 'c':
	  strcpy(in_clargs->proxy,optarg);
	  break;


      case 'U': /* grant option */
 	  passwd = getpwnam(optarg);
	  if(passwd) {
	      in_clargs->uid = passwd->pw_uid;
	  }      
	  break;

      case 'G': /* grant option */
	  group = getgrnam(optarg);
	  if(group) {
	      in_clargs->gid = group->gr_gid;
	  }      
	  break;
	  
      default:
	  return 1;
      }
  }
  
  dprintf
    ("Parsed Args: url=%s host=%s, path=%s, local_mountpoint=%s, amode=%s, port=%d\n",
     buf, in_clargs->host, in_clargs->path, in_clargs->mpoint,
		in_clargs->amode, in_clargs->port);
  return 0;
}

int usage(int argc, char **argv)
{
  fprintf(stderr, "%s: <url> <mountpoint> <options>...\n", argv[0]);
  return 0;
}				/* end of USAGE() */

/* do_mtab - Given a pointer to a filled struct mntent,
 * add an entry to /etc/mtab.
 *
 */
int do_mtab(struct mntent *myment)
{
  struct mntent *ment;
  FILE *mtab;
  FILE *tmp_mtab;

  mtab = setmntent(MTAB, "r");
  tmp_mtab = setmntent(TMP_MTAB, "w");

  if (mtab == NULL) {
    fprintf(stderr, "ERROR: couldn't open " MTAB " for read\n");
    endmntent(mtab);
    endmntent(tmp_mtab);
    return 1;
  } else if (tmp_mtab == NULL) {
    fprintf(stderr, "ERROR: couldn't open " TMP_MTAB " for write\n");
    endmntent(mtab);
    endmntent(tmp_mtab);
    return 1;
  }

  while ((ment = getmntent(mtab)) != NULL) {
    if (strcmp(myment->mnt_dir, ment->mnt_dir) != 0) {
      if (addmntent(tmp_mtab, ment) == 1) {
	fprintf(stderr, "ERROR: couldn't add entry to" TMP_MTAB "\n");
	endmntent(mtab);
	endmntent(tmp_mtab);
	return 1;
      }

    }
  }

  endmntent(mtab);

  if (addmntent(tmp_mtab, myment) == 1) {
    fprintf(stderr, "ERROR: couldn't add entry to " TMP_MTAB "\n");
    return 1;
  }

  endmntent(tmp_mtab);

  if (rename(TMP_MTAB, MTAB)) {
    fprintf(stderr, "ERROR: couldn't rename " TMP_MTAB " to " MTAB "\n");
    return 1;
  }

  return 0;
}				/* end of do_mtab */

/* setup_mntent - allocate space for members of mntent, fill in the data
 * structure
 */
int setup_mntent(struct mntent *myment, const char *fsname, const char *dir,
		 const char *type, const char *opts, int freq, int passno)
{
  myment->mnt_fsname = malloc(sizeof(char[strlen(fsname) + 1]));
  myment->mnt_dir = malloc(sizeof(char[strlen(dir) + 1]));
  myment->mnt_type = malloc(sizeof(char[strlen(type) + 1]));
  myment->mnt_opts = malloc(sizeof(char[strlen(opts) + 1]));

  if ((myment->mnt_fsname == NULL) || (myment->mnt_dir == NULL)
      || (myment->mnt_type == NULL) || (myment->mnt_opts == NULL)) {
    fprintf(stderr, "ERROR: cannot allocate memory\n");
    return 1;
  }

  strcpy(myment->mnt_fsname, fsname);
  strcpy(myment->mnt_dir, dir);
  strcpy(myment->mnt_type, type);
  strcpy(myment->mnt_opts, opts);
  myment->mnt_freq = freq;
  myment->mnt_passno = passno;

  return 0;
}				/* end setup_mntent */

int talktodavfsd(struct dav_mount_data *dav_data, char *url) {
    
    char buf[BUF_SIZE];
    char *pkg, *version;

    nsocket *ns = create_sock(dav_data->dav_fd);

    /* read Greeting */
    sock_readline(ns, buf, BUF_SIZE);
	
    pkg = strtok(buf, "\t");
    version = strtok(NULL, "\t");

    if(!pkg || !version || strcmp(pkg, PACKAGE) || strcmp(version, VERSION)) {
	fprintf(stderr, 
	"Diffrent version or davfs!! Please check davfs version or restart new davfsd");
	return -1;
    }

    /* send url */
    sock_sendline(ns,url);

    /* send proxy info */
    sock_sendline(ns,dav_data->proxy);

    /* send id */
    sock_sendline(ns,dav_data->id);	

    /* send pass */
    sock_sendline(ns,dav_data->pass);

    return 0;
}
/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 * End:
 */

int main(int argc, char **argv)
{
  int ret, flags = 0;
  char type[] = "davfs";
  struct dav_mount_data data;
  struct mntent myment;
  struct in_addr haddr;
  char url[DAV_HOSTLEN];

  /* initilize socket */
  sock_init();

  if (argc < 2) {
    fprintf(stderr, "%s: too few arguments\n", argv[0]);
    return (1);
  }

  /* Check if the ncpfs filesystem is in the kernel.  If not, attempt
   * to load the ncpfs module */
  if (load_davfs() != 0) {
    fprintf(stderr, "Error: Unable to load davfs, exiting...\n");
    exit(1);
  }

#ifndef AUTOFS
  strcpy(url, argv[1]);
  
  /* add end slash */
  if (*(url + strlen(url) - 1) != '/')
		  strcat(url, "/");
  
  memset(&data, 0, sizeof(data));
  
  if ((ret = parse_args(argc, argv, &data))) {
		  fprintf(stderr, "ERROR: problem parsing cmdline args\n");
		  return (ret);
  }
#else
  if (strstr( argv[0], "mount.davfs" ) ) {
		  /* We're being called by in the mount -tdavfs style */
		  if ((ret = parse_mount_args( argc, argv, &data, url ))) {
				  fprintf( stderr, "ERROR: problem parsing cmdline args\n" );
				  return(ret);
		  }
  }
  else {
		  strcpy(url, argv[1]);
		  /* add end slash */
		  if(*(url+strlen(url)-1)!='/')
				  strcat(url, "/");

		  if ((ret = parse_args(argc, argv, &data))) {
				  fprintf(stderr, "ERROR: problem parsing cmdline args\n");
				  return(ret);
		  }
  }
#endif

  /* check dav server */
  ret = check_dav_server(url, data.id, data.pass,data.proxy);

  /* get new passwd if get 404 and no id is specified */
  if (ret == 401 && data.id!=NULL) {
    char *pass;
    printf("Connected WEB-DAV server %s\n", url);
    printf("User Name: ");
    fgets(data.id, sizeof(data.id) - 1, stdin);
    strtok(data.id, "\n\r");
    printf("Password: ");
    strcpy(data.pass, "");
    pass = getpass(data.pass);
    if (pass)
      strncpy(data.pass, pass, sizeof(data.pass) - 1);

    /* add end slash */
    if (*(url + strlen(url) - 1) != '/')
      strcat(url, "/");

    ret = check_dav_server(url, data.id, data.pass, data.proxy);
  }

  if (ret != 200)
    return ret;

  if ((ret = ResolveAddr(data.host, &haddr))) {
    return (ret);
  }

  /* resolved address  for speed up */
  //strcpy(data.host,inet_ntoa(haddr));

  /* connect to local davfsd */
  data.dav_fd = connect_davfsd(DAVFSD_HOST, DAVFSD_PORT);
  if (data.dav_fd == -1) {
    /* try to run davfsd */
    run_davfsd();

    /* connect to local davfsd, again */
    data.dav_fd = connect_davfsd(DAVFSD_HOST, DAVFSD_PORT);
    if (data.dav_fd == -1) {
      perror("davfsd");
      return (ret);
    }
  }

	
  /* communicate with davfsfd */
  if(talktodavfsd(&data, url)==-1) 
      return -1;

  daemonize();

  dprintf("hostname after transform | hostname=%s\n", data.host);

  dprintf("calling mount(url = %s, mountpoint = %s, type = %s,"
	  " flags = %x, data = %x)\n", url, data.mpoint, type, flags,
	  (unsigned int) &data);

  if ((ret = mount(url, data.mpoint, type, flags, (void *) &data)) < 0) {
    perror("mount");
    close(data.dav_fd);
    return (ret);
  }

  if ((ret = setup_mntent(&myment, url, data.mpoint, type, data.amode, 0, 0))) {
    perror("mntent");
    close(data.dav_fd);
    return (ret);
  }

  if ((ret = do_mtab(&myment))) {
    perror("mtab");
    close(data.dav_fd);
    return (ret);
  }

  return (ret);
}				/* end of main() */

